import pandas as pd


def find_outliers(df):
    """
    Function that will remove the outliers present.
    """
    outliers = pd.DataFrame()
    for column in df.select_dtypes(include=['float64', 'int64']).columns:
        Q1 = df[column].quantile(0.25)
        Q3 = df[column].quantile(0.75)
        IQR = Q3 - Q1
        lower_bound = Q1 - 1.5 * IQR
        upper_bound = Q3 + 1.5 * IQR
        column_outliers = df[(df[column] < lower_bound) | (df[column] > upper_bound)]
        outliers = pd.concat([outliers, column_outliers])
    return outliers.drop_duplicates()

def split_item_id(df):
    """
    Function that will split the item id column into depart id and item id.
    """
# Check if 'item_id' column exists in the DataFrame
    if 'item_id' not in df.columns:
        raise ValueError("DataFrame must contain 'item_id' column.")

    # Split the 'item_id' column
    components = df['item_id'].str.split('_', expand=True)

    # Create new columns in the DataFrame
    df['depart_id'] = components[0]+ '_' + components[1]  # Category
    df['item_id'] = components[2]  # Identifier
    return df
def split_store_id(df):
    """
    Function that will split the store id column into state and store id.
    """
# Check if 'item_id' column exists in the DataFrame
    if 'store_id' not in df.columns:
        raise ValueError("DataFrame must contain 'store_id' column.")

    # Split the 'store_id' column
    components = df['store_id'].str.split('_', expand=True)

    # Create new columns in the DataFrame
    df['state'] = components[0]  # Category
    df['store_id'] = components[1]  # Identifier
    return df

def make_dataset_model(train_df,test_df):
  """
    Function that will perform all the above functions '
    to develop a training and testing dataset for modelling phase.

    returns: Training and Testing Data
  """
  train_df = find_outliers(train_df)
  print("Removed Outliers Training")
  test_df = find_outliers(test_df)
  print("Removed Outliers Testing")
  train_df = split_item_id(train_df)
  train_df=split_store_id(train_df)
  print("Spliting Item ID Outliers Training...")
  print("Spliting Store ID Outliers Training...")
  test_df = split_item_id(test_df)
  test_df=split_store_id(test_df)
  print("Spliting Item ID Outliers Testing...")
  print("Spliting Store ID Outliers Testing...")
  return train_df,test_df


import pandas as pd
import numpy as np


def test_data(train_df,test_df):
  """
    Preparing the dataframes for Training and Testing 
  """
  test_col=train_df.iloc[:,0:6]
  df=pd.concat([test_col,test_df],axis=1)
  col=["id","dept_id","state_id","cat_id"]
  train_df=train_df.drop(columns=col)
  test_df=df.drop(columns=col)
  return train_df,test_df

def melt_df(df):
    """
    Melting the "d" column to have a single column.

    """
    l=['item_id','store_id']
    df= pd.melt(df, id_vars=l, var_name='d',value_name='sold').dropna()
    return df

def merge_df(df,cal,items):
  """
    Merging the data frame to develop a final training set and testing test.
  """
  merge1=pd.merge(df,cal,on="d",how='left')
  merge2 =pd.merge(merge1,items,on=["store_id","item_id","wm_yr_wk"],how='left')
  df=merge2.drop(columns="d")
  return df

def make_train(df):
  """
    Building a Target column and returns a final Training dataset
  """
  df=df.dropna()
  df=df.reset_index(drop=True)
  df['revenue'] = df['sold']*df['sell_price']
  df=df.drop(columns=["sell_price","sold","wm_yr_wk"]).reset_index(drop=True)
  df.to_csv("train_df.csv",index=False)
  return df

def make_test(df):
  """
    Building a Target column and returns a final Testing dataset

  """
  df=df.dropna()
  df=df.reset_index(drop=True)
  df['revenue'] = df['sold']*df['sell_price']
  df=df.drop(columns=["sell_price","sold","wm_yr_wk"]).reset_index(drop=True)
  df.to_csv("test_df.csv",index=False)
  return df

def make_data_train_test(train_df,test_df,cal,items):
  """
    Function that will perform all the above functions to develop a merged dataset.

    returns: Training and Testing Data
  """
  train_df,test_df=test_data(train_df,test_df)
  train_melt=melt_df(train_df)
  test_melt=melt_df(test_df)
  train_df=merge_df(train_melt,cal,items)
  test_df=merge_df(test_melt,cal,items)
  train_df=make_train(train_df)
  test_df=make_test(test_df)
  return train_df,test_df

print("Starting with the Prediction Analysis........")


from sklearn.preprocessing import OneHotEncoder
from sklearn.model_selection import train_test_split
from sklearn.model_selection import RandomizedSearchCV
from sklearn.preprocessing import RobustScaler
from sklearn.metrics import  mean_squared_error, r2_score
import warnings
warnings.filterwarnings('ignore')
import pandas as pd
import numpy as np
import pickle

def ohe_encoding(x_train: pd.DataFrame, x_val: pd.DataFrame, cols: list):
    """
    Application of One Hot Encoding on training and validation sets.

    Parameters:
    - x_train: DataFrame containing training data
    - x_val: DataFrame containing validation data
    - cols: List of columns to apply One Hot Encoding to

    Returns:
    - x_train, x_val: DataFrames with One Hot Encoded columns added
    - ohe: Fitted OneHotEncoder instance
    """
    
    # Initialize the OneHotEncoder
    ohe = OneHotEncoder(sparse_output=False, handle_unknown='ignore')

    # Fit and Transform on Training Set
    ohe_train = ohe.fit_transform(x_train[cols])
    
    # Transform the Validation Set
    ohe_val = ohe.transform(x_val[cols])

    # Extracting feature names
    names = ohe.get_feature_names_out(cols)

    # Creating One Hot Encoded DataFrames
    ohe_train_df = pd.DataFrame(ohe_train, columns=names, index=x_train.index)
    ohe_val_df = pd.DataFrame(ohe_val, columns=names, index=x_val.index)

    # Concatenating the One Hot Encoded DataFrames with the original DataFrames
    x_train = pd.concat([x_train.drop(columns=cols), ohe_train_df], axis=1)
    x_val = pd.concat([x_val.drop(columns=cols), ohe_val_df], axis=1)

    return x_train, x_val, ohe

def encode_dates(df, date_col):

  df[date_col] = pd.to_datetime(df[date_col])
  df['year'] = df[date_col].dt.year
  df['month'] = df[date_col].dt.month
  df['day'] = df[date_col].dt.day
  df['weekday'] = df[date_col].dt.weekday  # Monday=0, Sunday=6
  df=df.drop(columns=date_col)
  return df

def train_model(train_df, model, model_name, target_col, param_grid=None):
    """
    Train a specified machine learning model.
    Hyperparameter tuning can be performed if a param_grid is provided.

    Args:
        train_df (pd.DataFrame): Input DataFrame containing training data.
        model: A machine learning model to be trained (e.g., Logistic Regression, Xboost,Gradient Boosting).
        model_name (str): The name of the model (used for saving the model,ohe,scaler).
        target_col (str): The name of the target column in the dataset.
        param_grid (dict): Hyperparameter grid for tuning using RandomizedSearchCV.

    Returns:
        model: Trained machine learning model (after tuning if param_grid is provided).
        ohe: One Hot Encoded Model.
        scaler: Robust Scaler
    """
    # Separate features and target
    x = train_df.drop(target_col, axis=1)
    y = train_df[target_col]
    x_train, x_val, y_train, y_val = train_test_split(x, y, test_size=0.2, random_state=42)
    print("Splitting into training and validation done")
    
    # Apply one-hot encoding to categorical columns
    x_train, x_val, ohe = ohe_encoding(x_train, x_val, ["depart_id", "state"])
    print("One Hot encoding for training and validation done")

    # Encode dates
    x_train = encode_dates(x_train, "date")
    x_val = encode_dates(x_val, "date")
    print("Encoding Dates for training and validation done")
    
    # Scale features
    scaler = RobustScaler()
    scaler.fit(x_train)
    x_train_scaled = scaler.transform(x_train)
    x_val_scaled = scaler.transform(x_val)
    print("Robust Scaling for training and validation done")

    # Save the OneHotEncoder
    file_ohe = f'ohe_{model_name}.pkl'
    with open(file_ohe, 'wb') as f:
        pickle.dump(ohe, f)
    print("Pickle File for OHE done")
    # Save the Scaler
    file_scaler = f'scaler.pkl'
    with open(file_scaler, 'wb') as f:
        pickle.dump(scaler, f)
    print("Pickle File for Scaler done")

    # If param_grid is provided, perform hyperparameter tuning
    if param_grid is None:
        model.fit(x_train_scaled, y_train)
    else:
        reg = RandomizedSearchCV(
            estimator=model,
            param_distributions=param_grid,
            n_iter=10,
            scoring='neg_mean_squared_error',
            cv=3,
            verbose=1,
            random_state=42,
            n_jobs=-1
        )
        reg.fit(x_train_scaled, y_train)
        model = reg  # Use the best model found
    print("Model Training Stage done")

    # Predictions
    y_pred_train = model.predict(x_train_scaled)
    y_pred_val = model.predict(x_val_scaled)
    print("\nPrediction Stage done")

    # Performance metrics: Training Set
    mse_train = mean_squared_error(y_train, y_pred_train)
    rmse_train = np.sqrt(mse_train)
    r2_train = r2_score(y_train, y_pred_train)

    # Performance metrics: Validation Set
    mse_val = mean_squared_error(y_val, y_pred_val)
    rmse_val = np.sqrt(mse_val)
    r2_val = r2_score(y_val, y_pred_val)

    # Print the metrics
    print("Training Performance:")
    print(f"RMSE: {rmse_train:.4f}")
    print(f"R-squared: {r2_train:.4f}")

    print("\nValidation Performance:")
    print(f"RMSE: {rmse_val:.4f}")
    print(f"R-squared: {r2_val:.4f}")

    # Save the trained model
    model_filepath = f'{model_name}_model.pkl'
    with open(model_filepath, 'wb') as f:
        pickle.dump(model, f)

    # Return the model, OHE,scaler
    return model, ohe,scaler


import pandas as pd
import pickle
from sklearn.metrics import mean_squared_error, r2_score
import numpy as np
from .make_dataset_model import  split_item_id,split_store_id,find_outliers

def encode_dates(df, date_col):

  df[date_col] = pd.to_datetime(df[date_col])
  df['year'] = df[date_col].dt.year
  df['month'] = df[date_col].dt.month
  df['day'] = df[date_col].dt.day
  df['weekday'] = df[date_col].dt.weekday  # Monday=0, Sunday=6
  df=df.drop(columns=date_col)
  return df


def load_models(model_filepath, ohe_filepath,scaler_filepath):
    """
    Load the trained model and OneHotEncoder from disk.

    Args:
        model_filepath (str): Path to the saved model file.
        ohe_filepath (str): Path to the saved OneHotEncoder file.
        scaler_filepath (str): Path to the saved OneHotEncoder file.

    Returns:
        model: Loaded model.
        ohe: Loaded OneHotEncoder.
        scaler: Loaded Robust Scaler
    """
    with open(model_filepath, 'rb') as f:
        model = pickle.load(f)
    with open(ohe_filepath, 'rb') as f:
        ohe = pickle.load(f)
    with open(scaler_filepath, 'rb') as f:
        scaler = pickle.load(f)
    return model, ohe, scaler

def predict(test_df, model, ohe,scaler):
    """
    Make predictions using the trained model and preprocessed test data.

    Args:
        test_df (DataFrame): DataFrame containing the test data.
        model: Trained model for prediction.
        ohe: Fitted OneHotEncoder instance.

    Returns:
        None
    """
    test_df=find_outliers(test_df)
    test_df=split_store_id(test_df)
    test_df=split_item_id(test_df)
    x_test = test_df.drop(columns=['revenue'])  # All columns except 'revenue'
    y_test = test_df['revenue']  # Specify 'revenue' as the target variable

    # Apply one-hot encoding to categorical columns
    x_test = ohe.transform(x_test[["depart_id", "state"]])
    x_test = pd.DataFrame(x_test, columns=ohe.get_feature_names_out(), index=test_df.index)
    x_test = pd.concat([test_df.drop(columns=["depart_id", "state", "revenue"]), x_test], axis=1)

    # Encode dates
    x_test = encode_dates(x_test, "date")

    # Scale features using the loaded scaler
    x_test_scaled = scaler.transform(x_test)

    # Calculate performance metrics for test set
    y_pred_test = model.predict(x_test_scaled)
    mse_test = mean_squared_error(y_test, y_pred_test)
    rmse_test = np.sqrt(mse_test)
    r2_test = r2_score(y_test, y_pred_test)

    # Print the metrics
    print("\nTesting Performance:")
    print(f"MSE: {mse_test:.4f}")
    print(f"RMSE: {rmse_test:.4f}")
    print(f"R-squared: {r2_test:.4f}")
